from ftw.table.interfaces import ICatalogTableSourceConfig
from zope import schema


class IBridgeCatalogTableSourceConfig(ICatalogTableSourceConfig):
    """A ``ftw.tabbedview`` / ``ftw.table`` table source config interface
    for creating tabs listing content from another client.
    """

    bridge_remote_client_id = schema.TextLine(
        title=u'Remote client id',
        description=u'The ftw.bridge client id of the remote client where '
        u'the catalog should be queried.')

    bridge_remove_path = schema.Bool(
        title=u'Remove path from query',
        description=u'Unless the path is explicitly defined it is not '
        u'useful to use generated paths, since those path start with the '
        u'site path of the querying client and not the one of the queried '
        u'client. Thus we remove the path filters from the catalog query '
        u'by default.',
        default=True)



from ftw.bridge.client.exceptions import BridgeConfigurationError
from ftw.bridge.client.interfaces import IBridgeConfig
from zope.interface import implements
import os


class BridgeConfig(object):
    implements(IBridgeConfig)

    def get_url(self):
        url = os.environ.get('bridge_url', None)
        if url is None:
            raise BridgeConfigurationError()
        return url

    def get_bridge_ips(self):
        ips = os.environ.get('bridge_ips', None)
        if ips is None:
            raise BridgeConfigurationError()

        return [ip.strip() for ip in ips.strip().split(',')]

    def get_client_id(self):
        id_ = os.environ.get('bridge_client_id', None)
        if id_ is None:
            raise BridgeConfigurationError()

        return id_.strip()



from ftw.bridge.client.portlets import watcher
from ftw.bridge.client.testing import ZCML_LAYER
from ftw.testing import MockTestCase
from plone.app.portlets.interfaces import IPortletTypeInterface
from plone.portlets.interfaces import IPortletManager
from plone.portlets.interfaces import IPortletRenderer
from zope.component import getMultiAdapter
from zope.component import queryUtility
from zope.component.interfaces import IFactory
from zope.interface import Interface
from zope.publisher.interfaces.browser import IBrowserView


class TestWatcherPortlet(MockTestCase):

    layer = ZCML_LAYER

    def test_portlet_type_registered(self):
        portlet_interface = queryUtility(
            IPortletTypeInterface, name='ftw.bridge.client.watcher_portlet')
        self.assertEqual(portlet_interface, watcher.IWatcherPortlet)

    def test_portlet_factory_registerd(self):
        factory = queryUtility(
            IFactory, name='ftw.bridge.client.watcher_portlet')
        self.assertNotEquals(factory, None)

    def test_renderer(self):
        context = self.create_dummy()
        request = self.stub_request()
        view = self.providing_stub(IBrowserView)
        manager = self.providing_stub(IPortletManager)
        assignment = watcher.Assignment('client-one', '@@watch')

        plone_portal_state = self.stub()
        self.mock_adapter(plone_portal_state, IBrowserView,
                          (Interface, Interface), name='plone_portal_state')
        self.expect(plone_portal_state(context, request)).result(
            plone_portal_state)
        self.expect(plone_portal_state.anonymous()).result(False)

        self.replay()

        renderer = getMultiAdapter(
            (context, request, view, manager, assignment), IPortletRenderer)
        self.assertNotEqual(renderer, None)



from Products.CMFCore.utils import getToolByName
from ftw.bridge.client.interfaces import PORTAL_URL_PLACEHOLDER
from ftw.bridge.client.testing import EXAMPLE_CONTENT_LAYER
from ftw.bridge.client.utils import get_brain_url
from ftw.bridge.client.utils import get_object_url
from ftw.bridge.client.utils import to_utf8_recursively
from unittest2 import TestCase


class TestUtils(TestCase):

    layer = EXAMPLE_CONTENT_LAYER

    def test_get_object_url(self):
        folder = self.layer['folder']
        self.assertEqual(
            get_object_url(folder),
            '%sfeed-folder' % PORTAL_URL_PLACEHOLDER)

    def test_get_brain_url(self):
        catalog = getToolByName(self.layer['portal'], 'portal_catalog')
        query = {'path': {'query': '/plone/feed-folder/page',
                          'depth': 0}}

        brains = catalog(query)
        self.assertEqual(len(brains), 1)

        self.assertEqual(
            get_brain_url(brains[0]),
            '%sfeed-folder/page' % PORTAL_URL_PLACEHOLDER)


class TestToUtf8Recursively(TestCase):

    def test_converts_strings(self):
        self.assertEquals(str, type(to_utf8_recursively(u'foo')))
        self.assertEquals(str, type(to_utf8_recursively('foo')))

    def test_keeps_integers(self):
        self.assertEquals(2, to_utf8_recursively(2))

    def test_converts_lists_recursively(self):
        self.assertEquals([str, str],
                          map(type, to_utf8_recursively([u'foo', 'bar'])))
        self.assertEquals(['foo', 'bar'],
                          to_utf8_recursively([u'foo', 'bar']))

    def test_converts_dicts_recursively(self):
        input = {u'foo': u'Foo',
                 'bar': 'Bar'}

        self.assertEquals([str, str],
                          map(type, to_utf8_recursively(input).keys()))

        self.assertEquals([str, str],
                          map(type, to_utf8_recursively(input).values()))

        # assertEquals does not make any type comparision.
        # We just test that the value is the same.
        self.assertEquals(input, to_utf8_recursively(input))



from ftw.bridge.client import hooks
from ftw.bridge.client.testing import PLONE_INTEGRATION_TESTING
from ftw.testing import MockTestCase
from Products.CMFCore.utils import getToolByName


class TestPASPluginSetupHandler(MockTestCase):

    layer = PLONE_INTEGRATION_TESTING

    def test_setuphandler(self):
        acl_users = getToolByName(self.layer['portal'], 'acl_users')
        self.assertFalse(hooks.PLUGIN_ID in acl_users.objectIds())

        hooks.install_pas_plugin(self.layer['portal'])
        self.assertTrue(hooks.PLUGIN_ID in acl_users.objectIds())

        # Does not fail on another call
        hooks.install_pas_plugin(self.layer['portal'])



from Acquisition import aq_inner
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from ftw.bridge.client import _
from plone.app.portlets.portlets import base
from plone.memoize.compress import xhtml_compress
from plone.portlets.interfaces import IPortletDataProvider
from zope import schema
from zope.component import getMultiAdapter
from zope.formlib import form
from zope.interface import implements


class IWatcherPortlet(IPortletDataProvider):

    # Remote client to call through the bridge for receiving the recently
    # modified feed.
    client_id = schema.TextLine(
        title=u'Client ID',
        required=True)

    # Path and view to call on the remote client.
    path = schema.TextLine(
        title=u'Path',
        required=True)


class Assignment(base.Assignment):

    implements(IWatcherPortlet)

    def __init__(self, client_id, path):
        self.client_id = client_id
        self.path = path

    @property
    def title(self):
        return _(u'title_watcher_portlet',
                 default='Recently modified')


class Renderer(base.Renderer):

    _template = ViewPageTemplateFile('templates/watcher.pt')

    def __init__(self, *args):
        base.Renderer.__init__(self, *args)

        context = aq_inner(self.context)
        portal_state = getMultiAdapter(
            (context, self.request),
            name=u'plone_portal_state')
        self.anonymous = portal_state.anonymous()

    @property
    def available(self):
        return not self.anonymous

    @property
    def title(self):
        return _(u'title_watcher_portlet',
                 default='Recently modified')

    def render(self):
        return xhtml_compress(self._template())


class AddForm(base.AddForm):

    form_fields = form.Fields(IWatcherPortlet)
    label = _(u'title_watcher_portlet',
              default='Recently modified')

    def create(self, data):
        return Assignment(**data)



from AccessControl.SecurityInfo import ClassSecurityInfo
from OFS.Cache import Cacheable
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.PluggableAuthService.interfaces import plugins
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from ftw.bridge.client.interfaces import IBridgeConfig
from zope.component import getUtility
from zope.interface import implements
from zope.publisher.browser import BrowserView
import logging


logger = logging.getLogger('ftw.bridge.client.plugin')


class AddBridgePlugin(BrowserView):

    template = ViewPageTemplateFile('templates/add-pas-plugin.pt')

    def __call__(self):

        if 'form.button.Add' in self.request.form:
            name = self.request.form.get('id')
            title = self.request.form.get('title')

            plugin = BridgePlugin(name, title)
            self.context.context[name] = plugin

            self.request.response.redirect(
                self.context.absolute_url() +
                '/manage_workspace?manage_tabs_message=Plugin+added.')

        else:
            return self.template()


class BridgePlugin(BasePlugin, Cacheable):
    """A PAS Plugin that authenticates requests from the bridge.
    """

    implements(plugins.IExtractionPlugin, plugins.IAuthenticationPlugin)

    meta_type = 'Bridge Authentication Plugin'
    security = ClassSecurityInfo()

    def __init__(self, plugin_id, title=None):
        self._setId(plugin_id)
        self.title = title

    security.declarePrivate('extractCredentials')

    def extractCredentials(self, request):
        """Extract Credentials of bridge requests.
        """

        principal = request.get_header('X-BRIDGE-AC', None)
        origin = request.get_header('X-BRIDGE-ORIGIN', None)

        creds = {}

        if origin and principal:
            ip = self._get_request_ip(request)

            creds['id'] = principal
            creds['login'] = principal
            creds['ip'] = ip

            logger.debug('extractCredentials has %r : %r : %r',
                         principal, origin, ip)

        return creds

    security.declarePrivate('authenticateCredentials')

    def authenticateCredentials(self, credentials):
        """Authenticate Credentials for bridge
        """

        extractor = credentials.get('extractor')
        if extractor != self.getId():
            logger.debug('authenticateCredentials: wrong extractor %r != %r',
                         extractor, self.getId())
            return None

        config = getUtility(IBridgeConfig)
        ip = credentials.get('ip')
        allowed_ips = config.get_bridge_ips()

        if ip not in allowed_ips:
            logger.debug('authenticateCredentials: unkown ip %r not in %r',
                         ip, allowed_ips)
            return None

        login = credentials.get('login')
        logger.debug('authenticateCredentials: authenticating %r',
                     login)
        return login, login

    security.declarePrivate('_get_request_ip')

    def _get_request_ip(self, request):
        """Returns the IP address of the host which sent the
        request initally.
        """
        ips = request.environ.get(
            'REMOTE_ADDR',
            request.environ.get('HTTP_X_FORWARDED_FOR'))

        if ips is None or not hasattr(ips, 'split'):
            return ips
        else:
            return ips.split(',')[0].strip()



from ftw.bridge.client.testing import FUNCTIONAL_TESTING
from ftw.bridge.client.tests.base import RequestAwareTestCase
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from plone.app.testing import setRoles
from plone.testing.z2 import Browser
import os.path
import transaction


class TestWatcherPortletAddForm(RequestAwareTestCase):

    layer = FUNCTIONAL_TESTING

    def setUp(self):
        super(RequestAwareTestCase, self).setUp()
        self.portal = self.layer['portal']
        setRoles(self.portal, TEST_USER_ID,
                 ['Manager', 'Member', 'Site Administrator'])

        transaction.commit()

        self.browser = Browser(self.layer['app'])
        self.browser.handleErrors = False
        self.browser.open('%s/login_form' % self.portal.absolute_url())
        self.browser.getControl(name='__ac_name').value = TEST_USER_NAME
        self.browser.getControl(
            name='__ac_password').value = TEST_USER_PASSWORD
        self.browser.getControl(name='submit').click()
        self.assertIn('You are now logged in', self.browser.contents)

    def tearDown(self):
        super(RequestAwareTestCase, self).tearDown()
        setRoles(self.portal, TEST_USER_ID, ['Member'])
        self.browser.open('%s/logout' % self.portal.absolute_url())
        transaction.commit()

    def test_portlet_addable(self):
        url = os.path.join(self.portal.absolute_url(),
                           '@@manage-portlets')
        self.browser.open(url)
        self.assertNotIn('You do not have sufficient privileges',
                         self.browser.contents)
        self.assertIn('Recently modified', self.browser.contents)

        self.assertEqual(
            self.browser.getControl('Recently modified', index=1).optionValue,
            '/++contextportlets++plone.rightcolumn/+/' +
            'ftw.bridge.client.watcher_portlet')

    def test_portlet_addform(self):
        self.assertEqual(
            self.portal.restrictedTraverse(
                '++contextportlets++plone.rightcolumn').keys(),
            [])

        url = os.path.join(self.portal.absolute_url(),
                           '++contextportlets++plone.rightcolumn',
                           '+/ftw.bridge.client.watcher_portlet')
        self.browser.open(url)

        self.assertNotIn('You do not have sufficient privileges',
                         self.browser.contents)
        self.assertIn('Client ID', self.browser.contents)

        self.browser.getControl('Client ID').value = 'foo'
        self.browser.getControl('Path').value = '@@bar?baz=1'
        self.browser.getControl('Save').click()

        mapping = self.portal.restrictedTraverse(
            '++contextportlets++plone.rightcolumn')
        self.assertEqual(
            mapping.keys(),
            ['title_watcher_portlet'])
        assignment = mapping.get('title_watcher_portlet')
        self.assertEqual(assignment.client_id, u'foo')
        self.assertEqual(assignment.path, u'@@bar?baz=1')



from AccessControl import SecurityManagement
from AccessControl.users import SimpleUser
from ZODB.POSException import ConflictError
from ftw.bridge.client.brain import BrainResultSet
from ftw.bridge.client.exceptions import MaintenanceError
from ftw.bridge.client.interfaces import IBrainRepresentation
from ftw.bridge.client.interfaces import IBridgeRequest
from ftw.bridge.client.interfaces import PORTAL_URL_PLACEHOLDER
from ftw.bridge.client.request import replace_placeholder_in_data
from ftw.bridge.client.testing import BRIDGE_CONFIG_LAYER
from ftw.bridge.client.tests.base import RequestAwareTestCase
from mocker import ANY
from unittest2 import TestCase
from zope.app.component.hooks import setSite
from zope.component import getGlobalSiteManager
from zope.component import queryUtility, getUtility
from zope.interface.verify import verifyClass
import json
import urllib2


class TestReplacingPlaceholder(TestCase):

    def test_replace_placeholder_in_data(self):
        data = {
            'foo': u'bar %s baz' % PORTAL_URL_PLACEHOLDER,
            'bar': 2,
            'baz': {
                'sub': PORTAL_URL_PLACEHOLDER},
            'barbaz': [PORTAL_URL_PLACEHOLDER],
            'foobar': (PORTAL_URL_PLACEHOLDER,)}

        replace_placeholder_in_data(data, 'THEURL')
        self.assertEquals(data, {
                'foo': u'bar THEURL baz',
                'bar': 2,
                'baz': {
                    'sub': 'THEURL'},
                'barbaz': ['THEURL'],
                'foobar': ['THEURL']})


class TestBridgeRequestUtility(RequestAwareTestCase):

    layer = BRIDGE_CONFIG_LAYER

    def setUp(self):
        RequestAwareTestCase.setUp(self)

        user = SimpleUser('john.doe', 'pw', [], [])
        SecurityManagement.newSecurityManager(object(), user)

    def tearDown(self):
        SecurityManagement.noSecurityManager()

    def test_component_is_registered(self):
        self.replay()
        utility = queryUtility(IBridgeRequest)
        self.assertNotEqual(utility, None)

    def test_implements_interface(self):
        self.replay()
        utility = getUtility(IBridgeRequest)
        klass = type(utility)

        self.assertTrue(IBridgeRequest.implementedBy(klass))
        verifyClass(IBridgeRequest, klass)

    def test_request_path(self):
        response = self._create_response()
        self._expect_request(
            url='http://bridge/proxy/target-client/path/to/@@something'
            ).result(response)

        self.replay()
        utility = getUtility(IBridgeRequest)

        self.assertEqual(
            utility('target-client', 'path/to/@@something'),
            response)

    def test_request_path_with_leading_slash(self):
        response = self._create_response()
        self._expect_request(
            url='http://bridge/proxy/target-client/@@something'
            ).result(response)

        self.replay()
        utility = getUtility(IBridgeRequest)

        self.assertEqual(
            utility('target-client', '/@@something'),
            response)

    def test_request_default_headers(self):
        response = self._create_response()
        self._expect_request(
            headers={'X-BRIDGE-ORIGIN': 'current-client',
                     'X-BRIDGE-AC': 'john.doe'}
            ).result(response)

        self.replay()
        utility = getUtility(IBridgeRequest)

        self.assertEqual(
            utility('target-client', '@@something2'),
            response)

    def test_request_custom_headers(self):
        response = self._create_response()
        self._expect_request(
            headers={'X-BRIDGE-ORIGIN': 'current-client',
                     'X-BRIDGE-AC': 'john.doe',
                     'X-CUSTOM-HEADER': 'some data'}
            ).result(response)

        self.replay()
        utility = getUtility(IBridgeRequest)

        headers = {'X-CUSTOM-HEADER': 'some data'}
        self.assertEqual(len(headers), 1)
        self.assertEqual(
            utility('target-client', '@@something',
                    headers={'X-CUSTOM-HEADER': 'some data'}),
            response)
        self.assertEqual(len(headers), 1)

    def test_passed_data(self):
        response = self._create_response()
        self._expect_request(data={'foo': 'bar'}).result(response)

        self.replay()
        utility = getUtility(IBridgeRequest)
        self.assertEqual(
            utility('target-client', '@@something',
                    data={'foo': 'bar'}),
            response)

    def test_maintenance_response_raises_exception(self):
        self._expect_request().throw(urllib2.HTTPError(
                'url', 503, 'Service Unavailable', None, None))

        self.replay()
        utility = getUtility(IBridgeRequest)

        with self.assertRaises(MaintenanceError):
            utility('target-client', '@@something')

    def test_get_json(self):
        response = self._create_response(raw='[{"foo": "bar"}]')
        self._expect_request(
            url='http://bridge/proxy/target-client/@@json-view',
            ).result(response)

        self.replay()
        utility = getUtility(IBridgeRequest)

        self.assertEqual(
            utility.get_json('target-client', '@@json-view'),
            [{'foo': 'bar'}])

    def test_search_catalog(self):
        response_data = [
            {'Title': 'Foo',
             '_url': 'https://target-client/foo',
             'portal_type': 'Folder'}]
        query = {'portal_type': ['Folder']}

        response = self._create_response(raw=json.dumps(response_data))
        response.headers = {'X-total_results_length': '1'}
        url = 'http://bridge/proxy/foo/@@bridge-search-catalog'

        self._expect_request(
            url=url,
            data={'query': json.dumps({'portal_type': ['Folder'],
                                       'batching_start': 0}),
                  'limit': 50}).result(response)

        self.replay()
        utility = getUtility(IBridgeRequest)
        results = utility.search_catalog('foo', query)

        self.assertEqual(type(results), BrainResultSet)
        self.assertEqual(len(results), 1)
        brain = results[0]
        self.assertTrue(IBrainRepresentation.providedBy(brain))
        self.assertEqual(brain.getURL(), 'https://target-client/foo')
        self.assertEqual(brain.portal_type, 'Folder')
        self.assertEqual(brain.Title, 'Foo')

    def test_get_json_error(self):
        self._expect_request().throw(
            urllib2.URLError('Connection failed'))

        site = self.stub()
        self.expect(site.getSiteManager()).call(getGlobalSiteManager)
        self.expect(site.error_log.raising(ANY))

        self.replay()

        utility = getUtility(IBridgeRequest)
        setSite(site)

        self.assertEqual(
            utility.get_json('target-client',
                             'path/to/@@something', silent=True),
            None)

    def test_traversing(self):
        site = self.stub()
        self.expect(site.getSiteManager()).call(getGlobalSiteManager)
        self.expect(site.absolute_url()).result('http://nohost/plone')

        request = self.create_dummy(
            form={'ori': 'formdata'},
            RESPONSE=self.create_dummy(
                headers={}))
        self.expect(site.REQUEST).result(request)

        assertion_data = self.create_dummy()

        def view_method():
            assertion_data.form = request.form.copy()
            return 'the url %s@@view should be replaced' % (
                PORTAL_URL_PLACEHOLDER)

        with self.mocker.order():
            # test 1
            self.expect(site.restrictedTraverse('baz/@@view')()).call(
                view_method)
            # test 2
            self.expect(site.restrictedTraverse('baz/@@view')()).throw(
                Exception('failed'))
            # test 3
            self.expect(site.restrictedTraverse('baz/@@view')()).throw(
                ConflictError())

        self.replay()

        utility = getUtility(IBridgeRequest)
        setSite(site)

        # test 1
        response = utility(
            'current-client',
            'baz/@@view?foo=bar&baz=%s' % PORTAL_URL_PLACEHOLDER,
            data={'url': PORTAL_URL_PLACEHOLDER})
        self.assertEqual(assertion_data.form, {
                'foo': 'bar',
                'url': 'http://nohost/plone/',
                'baz': 'http://nohost/plone/'})
        self.assertEqual(request.form, {'ori': 'formdata'})
        self.assertEqual(response.code, 200)
        self.assertEqual(
            response.read(),
            'the url http://nohost/plone/@@view should be replaced')

        # test 2
        with self.assertRaises(urllib2.HTTPError) as cm:
            response = utility('current-client', 'baz/@@view?foo=bar')
        response = cm.exception
        self.assertEqual(request.form, {'ori': 'formdata'})
        self.assertEqual(response.code, 500)
        self.assertIn(response.read(), 'failed')

        # test 3
        with self.assertRaises(ConflictError):
            utility('current-client', 'baz/@@view?foo=bar')

    def test_silent_error_is_logged(self):
        self._expect_request().throw(
            urllib2.URLError('Connection failed'))

        site = self.stub()
        self.expect(site.getSiteManager()).call(getGlobalSiteManager)
        self.expect(site.error_log.raising(ANY))

        self.replay()

        utility = getUtility(IBridgeRequest)
        setSite(site)

        self.assertEqual(
            utility('target-client', 'path/to/@@something', silent=True),
            None)

    def test_nonsilent_error_is_reraised(self):
        self._expect_request().throw(
            urllib2.URLError('Connection failed'))

        self.replay()
        utility = getUtility(IBridgeRequest)
        with self.assertRaises(urllib2.URLError):
            utility('target-client', 'path/to/@@something')



from AccessControl import SecurityManagement
from Products.CMFCore.utils import getToolByName
from ftw.bridge.client.portlets import watcher
from ftw.bridge.client.testing import INTEGRATION_TESTING
from plone.portlets.constants import USER_CATEGORY
from plone.portlets.interfaces import IPortletManager
from unittest2 import TestCase
from zope.component import getUtility
from zope.component import queryMultiAdapter


class TestAddWatcherPortletView(TestCase):

    layer = INTEGRATION_TESTING

    def _get_portlets(self, manager='plone.dashboard1'):
        context = self.layer['portal']
        column_manager = getUtility(IPortletManager, name=manager)
        membership_tool = getToolByName(context, 'portal_membership')
        userid = membership_tool.getAuthenticatedMember().getId()
        user_category = column_manager.get(USER_CATEGORY, None)
        if not user_category:
            return []

        column = user_category.get(userid, None)
        if not column:
            return []

        portlets = []
        for portlet in column.values():
            if watcher.IWatcherPortlet.providedBy(portlet):
                portlets.append(portlet)

        return portlets

    def test_component_is_registered(self):
        portal = self.layer['portal']
        request = self.layer['request']
        self.assertNotEquals(
            queryMultiAdapter((portal, request), name='add-watcher-portlet'),
            None)

    def test_add_portlet(self):
        portal = self.layer['portal']
        request = self.layer['request']

        request.environ['HTTP_X_BRIDGE_ORIGIN'] = 'client-one'
        request.form['path'] = '@@watcher-feed?uid=567891234'

        view = queryMultiAdapter((portal, request),
                                 name='add-watcher-portlet')

        self.assertEqual(len(self._get_portlets()), 0)

        self.assertEqual(view(), 'OK')
        portlets = self._get_portlets()
        self.assertEqual(len(portlets), 1)
        self.assertEqual(portlets[0].client_id, 'client-one')
        self.assertEqual(portlets[0].path, '@@watcher-feed?uid=567891234')

        self.assertEqual(view(), 'OK')
        self.assertEqual(len(self._get_portlets()), 2)

        self.assertEqual(view(), 'OK')
        self.assertEqual(len(self._get_portlets()), 3)

    def test_add_portlet_fails_with_anonymous(self):
        portal = self.layer['portal']
        request = self.layer['request']

        request.environ['HTTP_X_BRIDGE_ORIGIN'] = 'client-one'
        request.form['path'] = '@@watcher-feed?uid=567891234'

        sm = SecurityManagement.getSecurityManager()
        SecurityManagement.noSecurityManager()

        try:
            view = queryMultiAdapter((portal, request),
                                     name='add-watcher-portlet')
            with self.assertRaises(Exception) as cm:
                view()
            self.assertEqual(str(cm.exception), 'Could not find userid.')

        finally:
            SecurityManagement.setSecurityManager(sm)



from AccessControl import getSecurityManager
from StringIO import StringIO
from ZODB.POSException import ConflictError
from ftw.bridge.client.exceptions import MaintenanceError
from ftw.bridge.client.interfaces import IBrainSerializer
from ftw.bridge.client.interfaces import IBridgeConfig
from ftw.bridge.client.interfaces import IBridgeRequest
from ftw.bridge.client.interfaces import PORTAL_URL_PLACEHOLDER
from ftw.bridge.client.utils import json
from zope.app.component.hooks import getSite
from zope.component import getUtility
from zope.interface import implements
import os.path
import sys
import types
import urllib
import urllib2
import urlparse


def replace_placeholder_in_data(data, public_url):
    if isinstance(data, types.StringTypes):
        return data.replace(PORTAL_URL_PLACEHOLDER, public_url)

    elif isinstance(data, types.DictType):
        for key, value in data.items():
            data[key] = replace_placeholder_in_data(value, public_url)
        return data

    elif isinstance(data, (types.ListType, types.TupleType)):
        new = []
        for value in data:
            new.append(replace_placeholder_in_data(value, public_url))
        return new

    else:
        return data


class BridgeRequest(object):
    implements(IBridgeRequest)

    def __call__(self, target, path, headers=None, data=None, silent=False):
        """Makes a request to a remote client.
        """
        try:
            return self._request(target, path, headers, data)

        except (urllib2.HTTPError, urllib2.URLError), exc:
            if getattr(exc, 'code', None) == 503:
                raise MaintenanceError()

            elif silent:
                getSite().error_log.raising(sys.exc_info())
                return None

            else:
                raise

    def get_json(self, *args, **kwargs):
        """Makes a request to a JSON view on a remote client, return the
        converted python objects.
        """
        response = self(*args, **kwargs)
        if response is None:
            return None
        else:
            return json.loads(response.read())

    def search_catalog(self, target, query, limit=50, batching_start=0):
        path = '@@bridge-search-catalog'
        query['batching_start'] = batching_start
        data = {'query': json.dumps(query),
                'limit': limit}

        response = self(target, path, data=data)
        data = json.loads(response.read())
        total_length = int(response.headers.get(
                'X-total_results_length', '0'))

        serializer = getUtility(IBrainSerializer)
        return serializer.deserialize_brains(data, total_length)

    def _get_url(self, config, target, path):
        if path.startswith('/'):
            path = path[1:]
        return '/'.join((config.get_url(), target, path))

    def _get_headers(self, config, additional_headers):
        headers = {
            'X-BRIDGE-ORIGIN': config.get_client_id(),
            'X-BRIDGE-AC': self._get_current_userid()}

        if additional_headers:
            headers.update(additional_headers)

        return headers

    def _get_current_userid(self):
        return getSecurityManager().getUser().getId()

    def _request(self, target, path, headers, data):
        config = getUtility(IBridgeConfig)
        if config.get_client_id() == target:
            return self._do_traverse(path, headers, data)

        else:
            url = self._get_url(config, target, path)
            headers = self._get_headers(config, headers)
            return self._do_request(url, headers, data)

    def _do_request(self, url, headers, data):
        handler = urllib2.ProxyHandler({})
        opener = urllib2.build_opener(handler)

        if data:
            data = urllib.urlencode(data)

        request = urllib2.Request(url, data, headers)
        return opener.open(request)

    def _do_traverse(self, path, headers, data):
        portal = getSite()
        public_url = portal.absolute_url() + '/'

        parsed_path = urlparse.urlparse(path)
        form_data = dict(urlparse.parse_qsl(parsed_path.query))
        if data:
            form_data.update(data)
        replace_placeholder_in_data(form_data, public_url)

        request = portal.REQUEST
        # we need to back up the request data and set them new for the
        # view which is called with the same request (restrictedTraverse)
        ori_form = request.form
        ori_response_headers = request.RESPONSE.headers
        request.RESPONSE.headers = {}
        request.form = form_data

        response_url = os.path.join(portal.absolute_url(), parsed_path.path)
        response = None

        try:
            response_data = portal.restrictedTraverse(parsed_path.path)()

        except ConflictError:
            raise

        except Exception, exc:
            # restore the request
            request.form = ori_form
            request.RESPONSE.headers = ori_response_headers

            code = 500
            msg = str(exc)
            hdrs = {}
            fp = StringIO(msg)
            raise urllib2.HTTPError(response_url, code, msg, hdrs, fp)

        else:
            # restore the request
            response_headers = request.RESPONSE.headers
            request.form = ori_form
            request.RESPONSE.headers = ori_response_headers

            response_data = StringIO(response_data.replace(
                    PORTAL_URL_PLACEHOLDER, public_url))

            response = urllib.addinfourl(
                response_data, headers=response_headers,
                url=response_url, code=200)

        return response



from datetime import datetime
from DateTime import DateTime
from ftw.bridge.client.brain import BrainRepresentation
from ftw.bridge.client.brain import BrainResultSet
from ftw.bridge.client.brain import BrainSerializer
from ftw.bridge.client.interfaces import IBrainRepresentation
from ftw.bridge.client.interfaces import IBrainSerializer
from ftw.bridge.client.testing import INTEGRATION_TESTING
from ftw.builder import Builder
from ftw.builder import create
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from Products.CMFCore.utils import getToolByName
from time import time
from unittest2 import TestCase
from zope.component import queryUtility
from zope.interface.verify import verifyClass
import json


class TestBrainSerializer(TestCase):

    layer = INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        setRoles(self.portal, TEST_USER_ID, ['Contributor'])
        self.folder = create(Builder('folder').titled('Feed folder'))
        create(Builder('page').titled(u'The page with uml\xe4uts')
               .within(self.folder))

    def test_utility_is_registered(self):
        component = queryUtility(IBrainSerializer)
        self.assertNotEqual(None, component)

    def test_component_implements_interface(self):
        self.assertTrue(IBrainSerializer.implementedBy(
                BrainSerializer))
        verifyClass(IBrainSerializer, BrainSerializer)

    def test_serialize_brains(self):
        catalog = getToolByName(self.portal, 'portal_catalog')
        results = catalog(path='/'.join(self.folder.getPhysicalPath()),
                          sort_on='path')

        self.assertEqual(2, len(results))
        folder, page = results
        self.assertEquals('Folder', folder.portal_type)
        self.assertEquals('Document', page.portal_type)

        serializer = BrainSerializer()
        data = serializer.serialize_brains(results)
        self.assertEqual(list, type(data))
        self.assertEqual(2, len(data))

        folder_data, page_data = data

        self.assertEqual(u'***portal_url***feed-folder',
                         folder_data['_url'])
        self.assertEqual(u'Feed folder', folder_data['Title'])
        self.assertEqual(u'feed-folder', folder_data['id'])

        self.assertEqual(
            u'***portal_url***feed-folder/the-page-with-umlauts',
            page_data['_url'])
        self.assertEqual(u'The page with uml\xe4uts',
                         page_data['Title'])
        self.assertEqual(u'the-page-with-umlauts', page_data['id'])

    def test_serialized_data_is_jsonizable(self):
        catalog = getToolByName(self.portal, 'portal_catalog')
        results = catalog(path='/'.join(self.folder.getPhysicalPath()),
                          sort_on='path')

        self.assertEqual(2, len(results))

        serializer = BrainSerializer()
        data = serializer.serialize_brains(results)

        self.assertEqual(data, json.loads(json.dumps(data)))

    def test_deserializing(self):
        catalog = getToolByName(self.folder, 'portal_catalog')
        brains = catalog(path='/'.join(self.folder.getPhysicalPath()),
                          sort_on='path')

        self.assertEqual(2, len(brains))

        serializer = BrainSerializer()
        data = json.dumps(serializer.serialize_brains(brains))

        results = serializer.deserialize_brains(
            json.loads(data), total_length=13)

        self.assertEqual(BrainResultSet, type(results))
        self.assertEqual(2, len(results))
        self.assertEqual(13, results.get_total_length())
        folder, page = results

        self.assertTrue(IBrainRepresentation.providedBy(folder))
        self.assertTrue(IBrainRepresentation.providedBy(page))

        self.assertEqual('Feed folder', folder.Title)
        self.assertEqual('The page with uml\xc3\xa4uts', page.Title)

        # The url would be replaced by the bridge proxy, but here we
        # deserialize it directly.
        self.assertEqual('***portal_url***feed-folder',
                         folder.getURL())
        self.assertEqual(
            '***portal_url***feed-folder/the-page-with-umlauts',
            page.getURL())

        self.assertEqual('Folder', folder.portal_type)
        self.assertEqual('Document', page.portal_type)

    def test_converts_persistent_mutables_to_default_types(self):
        # On the remote side the data is not stored, therfore we can
        # convert persistent mutables to default types in order to be
        # able to dump it to json.

        serializer = BrainSerializer()
        self.assertEquals(dict, type(serializer._encode(PersistentMapping())),
                          'PersistentMapping should be converted to dict')
        self.assertEquals(list, type(serializer._encode(PersistentList())),
                          'PersistentList should be converted to list')

        self.assertEquals(dict, type(serializer._encode({'dict': PersistentMapping()})['dict']),
                          'Dicts should be encoded recursively')
        self.assertEquals(list, type(serializer._encode([PersistentList()])[0]),
                          'Lists should be encoded recursively')

    def test_converts_dates(self):
        serializer = BrainSerializer()
        data = {'python_datetime': datetime(2014, 12, 22, 15, 30),
                'zope_DateTime': DateTime('2001/11/14 13:22')}
        self.assertEquals(
            data,
            serializer._decode_data(json.loads(
                    json.dumps(serializer._encode(data)))))


class TestBrainResultSet(TestCase):

    def test_result_set(self):
        data = BrainResultSet(['foo', 'bar'], total_length=15)
        self.assertEqual(data[0], 'foo')
        self.assertEqual(data[1], 'bar')
        self.assertEqual(data.get_total_length(), 15)


class TestBrainRepresentation(TestCase):

    def test_component_implements_interface(self):
        self.assertTrue(IBrainRepresentation.implementedBy(
                BrainRepresentation))
        verifyClass(IBrainRepresentation, BrainRepresentation)

    def test_getURL(self):
        obj = BrainRepresentation({'_url': 'http://plone/foo'})
        self.assertEqual(obj.getURL(), 'http://plone/foo')

    def test_attribute_access(self):
        obj = BrainRepresentation({
                'Title': 'Foo',
                'portal_type': 'Folder'})

        self.assertEqual(obj.Title, 'Foo')
        self.assertEqual(obj.portal_type, 'Folder')

        with self.assertRaises(AttributeError) as cm:
            obj.missing_attr

        self.assertEqual(str(cm.exception), 'missing_attr')



from ftw.bridge.client.interfaces import IBridgeRequest
from ftw.bridge.client.tabbed import interfaces
from ftw.tabbedview.browser.listing import CatalogListingView
from ftw.table.catalog_source import CatalogTableSource
from zope.component import adapts
from zope.component import getUtility
from zope.interface import Interface
from zope.interface import implements


class BridgeCatalogTableSource(CatalogTableSource):
    """Custom catalog table source quering a remote client.
    """

    adapts(interfaces.IBridgeCatalogTableSourceConfig, Interface)

    def search_results(self, query):
        if self.config.bridge_remote_client_id is None:
            raise ValueError('%s defines no bridge_remote_client_id' % (
                    str(self.config)))

        query = self._remove_path_from_query(query)

        requester = getUtility(IBridgeRequest)
        client_id = self.config.bridge_remote_client_id
        pagesize = self.config.batching_pagesize
        current_page = self.config.batching_current_page

        results = requester.search_catalog(
            client_id, query, limit=pagesize,
            batching_start=((current_page - 1) * pagesize))

        return self._batch_results(results)

    def _batch_results(self, results):
        total_length = results.get_total_length()
        pagesize = self.config.pagesize
        page = self.config.batching_current_page - 1

        just_left = page * pagesize
        just_right = total_length - len(results) - just_left

        data = ([None] * just_left) + \
            results + \
            ([None] * just_right)
        return data

    def _remove_path_from_query(self, query):
        if self.config.bridge_remove_path and 'path' in query:
            del query['path']
        return query


class BridgeCatalogListingView(CatalogListingView):
    """A base ``ftw.tabbedview`` view acting as bridge catalog source config.
    """

    implements(interfaces.IBridgeCatalogTableSourceConfig)

    bridge_remote_client_id = None
    bridge_remove_path = True

    # Disable custom sort indexes since we have a lazy result.
    custom_sort_indexes = {}

    def custom_sort(self, results, sort_on, sort_reverse):
        return results



from AccessControl import SecurityManagement
from AccessControl.users import SimpleUser
from Products.statusmessages.interfaces import IStatusMessage
from ftw.bridge.client.testing import EXAMPLE_CONTENT_LAYER
from ftw.bridge.client.tests.base import RequestAwareTestCase
from plone.uuid.interfaces import IUUID
from zope.component import getMultiAdapter
from zope.component import queryMultiAdapter
import urllib2


class TestWatchActionView(RequestAwareTestCase):

    layer = EXAMPLE_CONTENT_LAYER

    def setUp(self):
        RequestAwareTestCase.setUp(self)

        user = SimpleUser('john.doe', 'pw', [], [])
        SecurityManagement.newSecurityManager(object(), user)

    def test_component_is_registered(self):
        context = self.stub()
        request = self.stub_request()

        self.replay()

        view = queryMultiAdapter((context, request), name='watch')
        self.assertNotEqual(view, None)

    def test_view_requests_bridge(self):
        context = self.layer['folder']
        request = self.layer['request']
        referer_url = 'http://nohost/plone/some-folder'
        request.environ['HTTP_REFERER'] = referer_url

        feed_path = '@@watcher-feed?uid=%s' % IUUID(context)
        bridge_path = 'http://bridge/proxy/dashboard/@@add-watcher-portlet'

        self._expect_request(url=bridge_path,
                             data={'path': feed_path}).result(
            self._create_response(raw='OK'))

        self.replay()

        view = getMultiAdapter((context, request), name='watch')
        view()
        self.assertEqual(request.response.headers.get('location'),
                         referer_url)

        messages = IStatusMessage(request).show()
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].message,
                         u'A dashboard portlet was created.')

    def test_maintenance(self):
        context = self.layer['folder']
        request = self.layer['request']
        referer_url = 'http://nohost/plone/some-folder'
        request.environ['HTTP_REFERER'] = referer_url

        self._expect_request().throw(urllib2.HTTPError(
                'url', 503, 'Service Unavailable', None, None))

        self.replay()

        view = getMultiAdapter((context, request), name='watch')
        view()
        self.assertEqual(request.response.headers.get('location'),
                         referer_url)

        messages = IStatusMessage(request).show()
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].message,
                         u'The target service is currently in ' + \
                             u'maintenace. Try again later.')

    def test_error(self):
        context = self.layer['folder']
        request = self.layer['request']

        self._expect_request().throw(urllib2.HTTPError(
                'url', 500, 'Internal Server Error', None, None))

        self.replay()

        view = getMultiAdapter((context, request), name='watch')
        view()
        self.assertEqual(request.response.headers.get('location'),
                         context.absolute_url())

        messages = IStatusMessage(request).show()
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].message,
                         u'The dashboard portlet could not be created.')

    def test_error_status_message_on_requests_exception(self):
        context = self.layer['folder']
        request = self.layer['request']

        self._expect_request().throw(
            urllib2.URLError('Connection failed'))

        self.replay()

        view = getMultiAdapter((context, request), name='watch')
        view()
        self.assertEqual(request.response.headers.get('location'),
                         context.absolute_url())

        messages = IStatusMessage(request).show()
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].message,
                         u'The dashboard portlet could not be created.')

    def test_redirect_to_file_view(self):
        # absolute_url would download the file, therefore we need to redirect
        # to the view - but we do that only for typesUseViewActionInListings
        context = self.layer['file']
        request = self.layer['request']

        self._expect_request().throw(
            urllib2.URLError('Connection failed'))

        self.replay()

        view = getMultiAdapter((context, request), name='watch')
        view()
        self.assertEqual(request.response.headers.get('location'),
                         context.absolute_url() + '/view')



from datetime import datetime
from ftw.bridge.client import _
from ftw.bridge.client.exceptions import MaintenanceError
from ftw.bridge.client.interfaces import IBridgeRequest
from ftw.bridge.client.interfaces import IWatcherPortletRegistry
from ftw.bridge.client.interfaces import MAINTENANCE_ERROR_MESSAGE
from ftw.bridge.client.portlets.watcher import Assignment
from ftw.bridge.client.utils import get_brain_url
from ftw.bridge.client.utils import get_object_url
from ftw.bridge.client.utils import json
from ftw.bridge.client.utils import to_utf8_recursively
from plone import api
from plone.app.portlets.utils import assignment_from_key
from plone.app.portlets.utils import assignment_mapping_from_key
from plone.portlets.constants import USER_CATEGORY
from plone.portlets.utils import unhashPortletInfo
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.PloneBatch import Batch
from Products.Five import BrowserView
from Products.statusmessages.interfaces import IStatusMessage
from zope.component import getMultiAdapter
from zope.component import getUtility
import time


WATCHER_PORTLET_LIMIT = 5
DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S'


class WatchAction(BrowserView):
    """Creates a "watcher" portlet on the client with the ID (or alias)
    "dashboard" through the bridge.
    """

    def __call__(self):
        uid = self.context.UID
        if callable(uid):
            uid = uid()

        feed_path = '@@watcher-feed?uid=%s' % uid

        requester = getUtility(IBridgeRequest)
        try:
            response = requester('dashboard', '@@add-watcher-portlet',
                                 data={'path': feed_path},
                                 silent=True)

        except MaintenanceError:
            IStatusMessage(self.request).addStatusMessage(
                MAINTENANCE_ERROR_MESSAGE, type='error')

        else:
            if response is None or response.code != 200:
                IStatusMessage(self.request).addStatusMessage(
                    _(u'error_portlet_creation_failed',
                      default=u'The dashboard portlet could not be created.'),
                    type='error')

            else:
                IStatusMessage(self.request).addStatusMessage(
                    _(u'info_msg_portlet_created',
                      default=u'A dashboard portlet was created.'),
                    type='info')

        referer = self.request.environ.get('HTTP_REFERER')
        if referer:
            self.request.RESPONSE.redirect(referer)

        else:
            state = getMultiAdapter((self.context, self.request),
                                    name='plone_context_state')
            self.request.RESPONSE.redirect(state.view_url())


class AddWatcherPortlet(BrowserView):

    def __call__(self):
        mapping = self._get_assignment_mapping()
        data = self._get_portlet_data()
        self._add_portlet(mapping, data)
        return 'OK'

    def _get_portlet_data(self):
        return {'client_id': self.request.get_header('X-BRIDGE-ORIGIN'),
                'path': self.request.get('path')}

    def _get_assignment_mapping(self):
        membership_tool = getToolByName(self.context, 'portal_membership')
        member = membership_tool.getAuthenticatedMember()

        if not member.getId():
            raise Exception(
                'Could not find userid.')

        userid = member.getId()

        return assignment_mapping_from_key(
            context=self.context,
            manager_name='plone.dashboard1',
            category=USER_CATEGORY,
            key=userid,
            create=True)

    def _add_portlet(self, mapping, data):
        portlet_id = self._generate_portlet_id(mapping)
        mapping[portlet_id] = Assignment(**data)
        mapping[portlet_id].__name__ = portlet_id
        return mapping[portlet_id]

    def _generate_portlet_id(self, column, base='watcher'):
        if base not in column:
            return base

        counter = 0
        while True:
            counter += 1
            id_ = base + str(counter)

            if id_ not in column:
                return id_


class WatcherFeed(BrowserView):

    def __call__(self):
        uid = self.request.get('uid')
        reference_catalog = getToolByName(self.context, 'reference_catalog')
        obj = reference_catalog.lookupObject(uid)
        data = self.get_data(obj)
        return json.dumps(data)

    def get_data(self, obj):
        return {
            'title': obj.Title().decode('utf-8'),
            'items': list(self.get_items(obj)),
            'details_url': '%s/@@watcher-recently-modified' % (
                get_object_url(obj))}

    def get_items(self, obj):
        types_to_ignore = api.portal.get_registry_record(
            name='types_to_ignore', interface=IWatcherPortletRegistry)

        brains = self.query_catalog(obj)
        for brain in brains:
            if types_to_ignore and brain.portal_type in types_to_ignore:
                continue

            yield self.get_item_data(brain)

    def query_catalog(self, obj, limit=WATCHER_PORTLET_LIMIT):
        catalog = getToolByName(self.context, 'portal_catalog')

        return catalog(path='/'.join(obj.getPhysicalPath()),
                       sort_on='modified',
                       sort_order='reverse',
                       sort_limit=limit)

    def get_item_data(self, brain):
        return {
            'title': to_utf8_recursively(brain.Title),
            'url': get_brain_url(brain),
            'modified': brain.modified.strftime(DATETIME_FORMAT),
            'portal_type': brain.portal_type,
            'cssclass': u'',
            }


class AjaxLoadPortletData(BrowserView):

    def __call__(self):
        portlet = self._get_portlet()
        try:
            data = self._get_data(portlet)
        except MaintenanceError:
            return '"MAINTENANCE"'

        data = self._localize_dates(data)
        return json.dumps(data)

    def _get_portlet(self):
        portlet_hash = self.request.get('hash')

        info = unhashPortletInfo(portlet_hash)

        return assignment_from_key(
            context=self.context,
            manager_name=info['manager'],
            category=info['category'],
            key=info['key'],
            name=info['name'])

    def _get_data(self, portlet):
        requester = getUtility(IBridgeRequest)
        return requester.get_json(portlet.client_id, portlet.path)

    def _localize_dates(self, data):
        translation = getToolByName(self.context, 'translation_service')
        localize_time = translation.ulocalized_time

        for item in data.get('items'):
            if item.get('modified'):
                date = datetime(*(time.strptime(
                            item['modified'], DATETIME_FORMAT)[0:6]))
                item['modified'] = localize_time(date, long_format=False)

        return data


class RecentlyModified(BrowserView):
    """A local recently modified view.
    """

    def __init__(self, *args, **kwargs):
        super(RecentlyModified, self).__init__(*args, **kwargs)
        self.batched_results = None

    def __call__(self):
        self.update()
        return super(RecentlyModified, self).__call__()

    def update(self):
        brains = self.query_catalog()
        self.batched_results = self.batch(brains)

    def query_catalog(self):
        catalog = getToolByName(self.context, 'portal_catalog')
        query = {
            'sort_on': 'modified',
            'sort_order': 'reverse',
            'path': '/'.join(self.context.getPhysicalPath())}

        return catalog(query)

    def batch(self, brains):
        b_start = self.request.get('b_start', 0)
        return Batch(brains, 20, start=int(b_start))



from StringIO import StringIO
from ftw.testing import MockTestCase
from mocker import ANY, ARGS, KWARGS
import urllib


class RequestAwareTestCase(MockTestCase):

    def setUp(self):
        MockTestCase.setUp(self)

        self.urllib2 = self.mocker.replace('urllib2')
        # Let the urllib2 not do any requests at all while
        # testing. We do this by expecting any request call zero times.
        self.expect(self.urllib2.build_opener(ARGS, KWARGS)).count(0)
        self.expect(self.urllib2.Request(ARGS, KWARGS)).count(0)

    def _create_response(self, status_code=200, raw='response data',
                         total_length=None):
        data = StringIO(raw)
        response = self.create_dummy(
            code=status_code,
            fp=data,
            read=data.read,
            headers={})
        return response

    def _expect_request(self, url=ANY, headers=ANY, data=None):
        data = data and urllib.urlencode(data) or ANY
        request = self.mocker.mock()
        self.expect(self.urllib2.Request(
                url,
                data,
                headers)).result(request)
        opener = self.mocker.mock()
        self.expect(self.urllib2.build_opener(ANY)).result(opener)
        return self.expect(opener.open(request))



from Products.CMFCore.utils import getToolByName
from StringIO import StringIO
from ftw.builder.testing import BUILDER_LAYER
from ftw.builder.testing import functional_session_factory
from ftw.builder.testing import set_builder_session_factory
from ftw.testing.layer import ComponentRegistryLayer
from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import PLONE_FUNCTIONAL_TESTING
from plone.app.testing import PLONE_INTEGRATION_TESTING
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import TEST_USER_ID
from plone.app.testing import applyProfile
from plone.app.testing import setRoles
from plone.testing import Layer
import os


CONFIG_VARIABLES = {
        'bridge_url': 'http://bridge/proxy',
        'bridge_ipds': '127.0.0.1, 127.0.0.2',
        'bridge_client_id': 'current-client',
        }


def setup_config():
    for key, value in CONFIG_VARIABLES.items():
        os.environ[key] = value


def teardown_config():
    for key in CONFIG_VARIABLES.keys():
        if key in os.environ:
            del os.environ[key]


class ZCMLLayer(ComponentRegistryLayer):
    """A layer which only sets up the zcml, but does not start a zope
    instance.
    """

    def setUp(self):
        super(ZCMLLayer, self).setUp()

        import ftw.bridge.client.tests
        self.load_zcml_file('tests.zcml', ftw.bridge.client.tests)


ZCML_LAYER = ZCMLLayer()


class BridgeConfigLayer(Layer):

    defaultBases = (ZCML_LAYER,)

    def setUp(self):
        setup_config()

    def tearDown(self):
        teardown_config()


BRIDGE_CONFIG_LAYER = BridgeConfigLayer()


class PloneTestingLayer(PloneSandboxLayer):

    defaultBases = (PLONE_INTEGRATION_TESTING,)

    def setUpZope(self, app, configurationContext):
        # Load ZCML
        import ftw.bridge.client
        self.loadZCML(package=ftw.bridge.client)


PLONE_TESTING_LAYER = PloneTestingLayer()


class IntegrationTestingLayer(PloneTestingLayer):

    defaultBases = (PLONE_INTEGRATION_TESTING,
                    BUILDER_LAYER)

    def setUpZope(self, app, configurationContext):
        # Load ZCML
        import ftw.bridge.client
        self.loadZCML(package=ftw.bridge.client)

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'ftw.bridge.client:default')

        mtool = getToolByName(portal, 'portal_membership')
        setRoles(portal, TEST_USER_ID, ['Manager'])
        portal.invokeFactory('Folder', 'Members', title='Members')
        mtool.setMemberareaCreationFlag()
        mtool.createMemberarea(TEST_USER_ID)
        setRoles(portal, TEST_USER_ID, ['Member'])


INTEGRATION_FIXTURE = IntegrationTestingLayer()
INTEGRATION_TESTING = IntegrationTesting(
    bases=(INTEGRATION_FIXTURE,), name='ftw.bridge.client:Integration')


class FunctionTestingLayer(PloneTestingLayer):

    defaultBases = (PLONE_FUNCTIONAL_TESTING, INTEGRATION_FIXTURE)


FUNCTIONAL_FIXTURE = FunctionTestingLayer()
FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(FUNCTIONAL_FIXTURE,
           set_builder_session_factory(functional_session_factory)),
    name='ftw.bridge.client:Functional')


class ExampleContentLayer(PloneSandboxLayer):

    defaultBases = (INTEGRATION_TESTING,)

    def setUpPloneSite(self, portal):
        setup_config()

        setRoles(portal, TEST_USER_ID, ['Manager'])
        self['folder'] = folder = portal.get(portal.invokeFactory(
                'Folder', 'feed-folder', title='Feed folder'))
        folder.invokeFactory('Document', 'page',
                             title='The page with uml\xc3\xa4uts')

        filedata = StringIO('foo bar')
        self['file'] = portal.get(
            portal.invokeFactory('File', 'file', file=filedata,
                                 title='the file'))

        setRoles(portal, TEST_USER_ID, ['Member'])

    def tearDownPloneSite(self, portal):
        teardown_config()


EXAMPLE_CONTENT_LAYER = ExampleContentLayer()



from Products.CMFCore.utils import getToolByName
from Products.PluggableAuthService.interfaces import plugins
from ftw.bridge.client.plugin import BridgePlugin


PLUGIN_ID = 'ftw-bridge'


def installed(site):
    install_pas_plugin(site)


def uninstalled(site):
    remove_pas_plugin(site)


def install_pas_plugin(site):
    acl_users = getToolByName(site, 'acl_users')

    if PLUGIN_ID in acl_users.objectIds():
        return

    plugin = BridgePlugin(PLUGIN_ID)
    acl_users._setObject(plugin.getId(), plugin)
    # acl_users[PLUGIN_ID] = plugin
    plugin = acl_users.get(PLUGIN_ID)

    # activate PAS plugin handlers
    plugin_interfaces = [plugins.IAuthenticationPlugin.__name__,
                         plugins.IExtractionPlugin.__name__]
    plugin.manage_activateInterfaces(plugin_interfaces)


def remove_pas_plugin(site):
    acl_users = getToolByName(site, 'acl_users')

    if PLUGIN_ID not in acl_users.objectIds():
        return

    del acl_users[PLUGIN_ID]



class BridgeConfigurationError(Exception):
    """The bridge was not configured well.
    """


class MaintenanceError(Exception):
    """The request to the remote client failed because the remote client is
    in maintenance mode.
    """



from ftw.bridge.client.exceptions import BridgeConfigurationError
from ftw.bridge.client.interfaces import IBridgeConfig
from ftw.bridge.client.testing import ZCML_LAYER
from ftw.testing import MockTestCase
from zope.component import queryUtility, getUtility
from zope.interface.verify import verifyClass
import os


class TestConfig(MockTestCase):

    layer = ZCML_LAYER

    def test_component_is_registered(self):
        config = queryUtility(IBridgeConfig)
        self.assertNotEqual(config, None)

    def test_implement_interface(self):
        config = getUtility(IBridgeConfig)
        klass = type(config)

        self.assertTrue(IBridgeConfig.implementedBy(klass))
        verifyClass(IBridgeConfig, klass)

    def test_get_url(self):
        os.environ['bridge_url'] = 'http://localhost:8000/proxy'
        config = queryUtility(IBridgeConfig)
        self.assertEqual(config.get_url(), 'http://localhost:8000/proxy')

    def test_get_url_without_config(self):
        if 'bridge_url' in os.environ:
            del os.environ['bridge_url']

        config = queryUtility(IBridgeConfig)
        with self.assertRaises(BridgeConfigurationError):
            config.get_url()

    def test_get_ips(self):
        os.environ['bridge_ips'] = '127.0.0.1, 127.0.0.2'
        config = queryUtility(IBridgeConfig)
        self.assertEqual(config.get_bridge_ips(), ['127.0.0.1', '127.0.0.2'])

    def test_get_ips_without_config(self):
        if 'bridge_ips' in os.environ:
            del os.environ['bridge_ips']

        config = queryUtility(IBridgeConfig)
        with self.assertRaises(BridgeConfigurationError):
            config.get_bridge_ips()

    def test_get_client_id(self):
        os.environ['bridge_client_id'] = 'client-one'
        config = getUtility(IBridgeConfig)
        self.assertEqual(config.get_client_id(), 'client-one')

    def test_get_client_id_without_config(self):
        if 'bridge_client_id' in os.environ:
            del os.environ['bridge_client_id']

        config = getUtility(IBridgeConfig)
        with self.assertRaises(BridgeConfigurationError):
            config.get_client_id()



from datetime import datetime
from ftw.bridge.client.interfaces import IBrainRepresentation
from ftw.bridge.client.interfaces import IBrainSerializer
from ftw.bridge.client.utils import get_brain_url
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping
from Products.CMFCore.utils import getToolByName
from zope.component.hooks import getSite
from zope.interface import implements
import DateTime
import Missing


class BrainSerializer(object):
    implements(IBrainSerializer)

    def serialize_brains(self, results):
        data = []
        for brain in results:
            data.append(self._serialize_brain(brain))
        return data

    def deserialize_brains(self, data, total_length=None):
        if not isinstance(data, list):
            raise ValueError('data should be a list, got %s' % type(data))

        results = []
        for brain_data in data:
            brain_data = self._decode_data(brain_data)
            results.append(BrainRepresentation(brain_data))

        return BrainResultSet(results, total_length)

    def _serialize_brain(self, brain):
        data = {self._encode('_url'): self._encode(get_brain_url(brain))}

        for name in self._get_metadata_names(brain):
            value = getattr(brain, name, None)
            data[self._encode(name)] = self._encode(value)

        return data

    def _encode(self, value):
        if isinstance(value, str):
            return value.decode('utf-8')

        elif value is Missing.Value:
            return [':Missing.Value']

        elif isinstance(value, DateTime.DateTime):
            return [':DateTime', str(value)]

        elif isinstance(value, datetime):
            return [':datetime', self._encode(value.timetuple()[:-2])]

        elif isinstance(value, tuple):
            return self._encode(list(value))

        elif isinstance(value, PersistentMapping):
            return self._encode(dict(value))

        elif isinstance(value, dict):
            return dict((self._encode(key), self._encode(value))
                        for key, value in value.items())

        elif isinstance(value, PersistentList):
            return self._encode(list(value))

        elif isinstance(value, list):
            return map(self._encode, value)

        return value

    def _decode(self, value):
        if isinstance(value, unicode):
            return value.encode('utf-8')

        elif isinstance(value, list) and len(value) < 1:
            return value

        elif isinstance(value, list) and value[0] == ':Missing.Value':
            return Missing.Value

        elif isinstance(value, list) and value[0] == ':DateTime':
            try:
                return DateTime.DateTime(value[1])
            except DateTime.interfaces.SyntaxError:
                return None

        elif isinstance(value, list) and value[0] == ':datetime':
            return datetime(*value[1])

        return value

    def _decode_data(self, data):
        new_data = {}
        for key, value in data.items():
            new_data[self._decode(key)] = self._decode(value)
        return new_data

    def _get_metadata_names(self, brain):
        catalog = getToolByName(getSite(), 'portal_catalog')
        return catalog._catalog.names


class BrainResultSet(list):

    def __init__(self, data, total_length):
        super(BrainResultSet, self).__init__(data)
        self._total_length = total_length

    def get_total_length(self):
        return self._total_length


class BrainRepresentation(object):
    implements(IBrainRepresentation)

    def __init__(self, data):
        self._data = data

    def __getattr__(self, name):
        if name in self._data:
            return self._data[name]
        else:
            raise AttributeError(name)

    def getURL(self):
        return self._url



from AccessControl import SecurityManagement
from AccessControl.users import SimpleUser
from Products.CMFCore.utils import getToolByName
from Products.statusmessages.interfaces import IStatusMessage
from ftw.bridge.client.interfaces import PORTAL_URL_PLACEHOLDER
from ftw.bridge.client.testing import EXAMPLE_CONTENT_LAYER
from ftw.bridge.client.testing import INTEGRATION_TESTING
from ftw.bridge.client.tests.base import RequestAwareTestCase
from plone.mocktestcase.dummy import Dummy
from unittest2 import TestCase
from zope.component import getMultiAdapter
import urllib2


class TestAddWatcherPortletView(TestCase):

    layer = INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        self.request = self.layer['request']
        self.mtool = getToolByName(self.portal, 'portal_membership')
        self.folder = self.mtool.getHomeFolder()

    def tearDown(self):
        if 'title' in self.request.form:
            del self.request.form['title']

        if 'url' in self.request.form:
            del self.request.form['url']

        if 'Favorites' in self.folder.objectIds():
            self.folder.manage_delObjects(['Favorites'])

    def test_component_is_registered(self):
        self.portal.restrictedTraverse('@@add-favorite')

    def test_get_favorites_folder(self):
        self.assertNotIn('Favorites', self.folder.objectIds())
        view = self.portal.restrictedTraverse('@@add-favorite')
        folder = view.get_favorites_folder()
        self.assertNotEquals(folder, None)
        self.assertIn('Favorites', self.folder.objectIds())

    def test_adds_link_object(self):
        self.assertNotIn('Favorites', self.folder.objectIds())
        view = self.portal.restrictedTraverse('@@add-favorite')

        self.request.form['title'] = 'Google search'
        self.request.form['url'] = 'http://www.google.com/'

        self.assertEquals(view(), 'OK')
        folder = view.get_favorites_folder()
        self.assertEquals(len(folder.objectValues()), 1)
        link = folder.objectValues()[0]
        self.assertEquals(link.Title(), 'Google search')
        self.assertEquals(link.getRemoteUrl(), 'http://www.google.com/')

    def test_generate_favorite_id(self):
        mock_folder = Dummy(objectIds=lambda: ['favorite-1', 'favorite-2'])
        view = self.portal.restrictedTraverse('@@add-favorite')
        self.assertEqual(view.generate_favorite_id(mock_folder),
                         'favorite-3')


class TestRemoteAddFavoriteAction(RequestAwareTestCase):

    layer = EXAMPLE_CONTENT_LAYER

    def setUp(self):
        RequestAwareTestCase.setUp(self)

        user = SimpleUser('john.doe', 'pw', [], [])
        SecurityManagement.newSecurityManager(object(), user)

        self.portal = self.layer['portal']
        self.request = self.layer['request']
        self.page = self.portal.get('feed-folder').get('page')
        self.referer_url = 'http://nohost/plone/feed-folder/page'
        self.request.environ['HTTP_REFERER'] = self.referer_url

    def tearDown(self):
        IStatusMessage(self.request).show()
        if 'HTTP_REFERER' in self.request.environ:
            del self.request.environ['HTTP_REFERER']
        SecurityManagement.noSecurityManager()

        if 'location' in self.request.response.headers:
            del self.request.response.headers['location']

    def test_component_is_registered(self):
        self.replay()
        getMultiAdapter((self.page, self.request),
                        name='remote-add-favorite')

    def test_requests_bridge(self):
        favorite_url = '%sfeed-folder/page' % PORTAL_URL_PLACEHOLDER
        bridge_url = 'http://bridge/proxy/dashboard/@@add-favorite'

        self._expect_request(url=bridge_url,
                             data={'title': 'The page with uml\xc3\xa4uts',
                                   'url': favorite_url}).result(
            self._create_response(raw='OK'))

        self.replay()
        view = getMultiAdapter((self.page, self.request),
                               name='remote-add-favorite')
        view()

    def test_status_message_on_success(self):
        self._expect_request().result(self._create_response(raw='OK'))

        self.replay()
        view = getMultiAdapter((self.page, self.request),
                               name='remote-add-favorite')
        view()

        messages = IStatusMessage(self.request).show()
        self.assertEqual(len(messages), 1)
        self.assertEqual(
            messages[0].message,
            u'The page with uml\xe4uts was added to your favorites.')
        self.assertEqual(messages[0].type,
                         'info')

    def test_redirects_back(self):
        self._expect_request().result(self._create_response(raw='OK'))

        self.replay()
        view = getMultiAdapter((self.page, self.request),
                               name='remote-add-favorite')
        view()

        self.assertEqual(self.request.response.headers.get('location'),
                         self.page.absolute_url())

    def test_maintenance_status_message(self):
        self._expect_request().throw(urllib2.HTTPError(
                'url', 503, 'Service Unavailable', None, None))

        self.replay()
        view = getMultiAdapter((self.page, self.request),
                               name='remote-add-favorite')
        view()

        messages = IStatusMessage(self.request).show()
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].message,
                         u'The target service is currently in ' + \
                             u'maintenace. Try again later.')
        self.assertEqual(messages[0].type,
                         'error')

    def test_maintenance_redirect(self):
        self._expect_request().throw(urllib2.HTTPError(
                'url', 503, 'Service Unavailable', None, None))

        self.replay()
        view = getMultiAdapter((self.page, self.request),
                               name='remote-add-favorite')
        view()

        self.assertEqual(self.request.response.headers.get('location'),
                         self.page.absolute_url())

    def test_error_status_message(self):
        self._expect_request().result(
            self._create_response(status_code=500, raw='Error'))
        self.replay()

        view = getMultiAdapter((self.page, self.request),
                               name='remote-add-favorite')
        view()

        messages = IStatusMessage(self.request).show()
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].message,
                         u'The favorite could not be created.')
        self.assertEqual(messages[0].type,
                         'error')

    def test_error_redirect(self):
        self._expect_request().result(
            self._create_response(status_code=500, raw='Error'))

        self.replay()
        view = getMultiAdapter((self.page, self.request),
                               name='remote-add-favorite')
        view()

        self.assertEqual(self.request.response.headers.get('location'),
                         self.page.absolute_url())

    def test_redirects_to_context_when_no_referer_is_found(self):
        if 'HTTP_REFERER' in self.request.environ:
            del self.request.environ['HTTP_REFERER']
        self._expect_request().result(
            self._create_response(status_code=500, raw='Error'))

        self.replay()
        view = getMultiAdapter((self.page, self.request),
                               name='remote-add-favorite')
        view()

        self.assertEqual(self.request.response.headers.get('location'),
                         self.page.absolute_url())

    def test_error_status_message_on_requests_exception(self):
        self._expect_request().throw(
            urllib2.URLError('Connection failed'))

        self.replay()

        view = getMultiAdapter((self.page, self.request),
                               name='remote-add-favorite')
        view()

        messages = IStatusMessage(self.request).show()
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].message,
                         u'The favorite could not be created.')
        self.assertEqual(messages[0].type,
                         'error')

    def test_redirect_to_file_view(self):
        # absolute_url would download the file, therefore we need to redirect
        # to the view - but we do that only for typesUseViewActionInListings
        self._expect_request().throw(
            urllib2.URLError('Connection failed'))

        self.replay()

        if 'HTTP_REFERER' in self.request.environ:
            del self.request.environ['HTTP_REFERER']

        context = self.portal.get('file')
        view = getMultiAdapter((context, self.request),
                               name='remote-add-favorite')
        view()
        self.assertEqual(self.request.response.headers.get('location'),
                         context.absolute_url() + '/view')

    def test_file_url_has_view_in_favoritte(self):
        # absolute_url would download the file, therefore we need to set the
        # favorite to point to the view - but we do that only for
        # typesUseViewActionInListings.

        favorite_url = '%sfile/view' % PORTAL_URL_PLACEHOLDER
        bridge_url = 'http://bridge/proxy/dashboard/@@add-favorite'

        self._expect_request(url=bridge_url,
                             data={'title': 'the file',
                                   'url': favorite_url}).result(
            self._create_response(raw='OK'))

        self.replay()
        context = self.portal.get('file')
        view = getMultiAdapter((context, self.request),
                               name='remote-add-favorite')
        view()



from Products.CMFCore.utils import getToolByName
from Products.Five import BrowserView
from copy import deepcopy
from ftw.bridge.client.interfaces import IBrainSerializer
from ftw.bridge.client.utils import json
from ftw.bridge.client.utils import to_utf8_recursively
from zope.component import getUtility


class BridgeSearchCatalog(BrowserView):

    def __call__(self):
        query = to_utf8_recursively(json.loads(self.request.get('query')))

        # ftw.solr hack https://github.com/4teamwork/ftw.solr/issues/42
        if query.get('path') == '/' or (
            isinstance(query.get('path'), dict) and query['path'].get('query') == '/'):
            del query['path']

        limit = int(self.request.get('limit'))
        brains = self._query_catalog(query, limit)

        total_length = self._count_unbatched_length(query)
        response = self.request.RESPONSE 
        response.setHeader(
            'X-total_results_length', str(total_length))
        response.setHeader('X-Theme-Disabled', 'True')
        response.setHeader("Content-type", "application/json")

        return self._serialize_results(brains)

    def _query_catalog(self, query, limit):
        catalog = getToolByName(self.context, 'portal_catalog')

        if 'batching_start' in query:
            batching_start = int(query['batching_start'])
            del query['batching_start']
        else:
            batching_start = 0

        # ftw.solr may destroy our query
        # https://github.com/4teamwork/ftw.solr/issues/41
        query = deepcopy(query)
        brains = catalog(query)
        batching_stop = batching_start + limit

        return brains[batching_start: batching_stop]

    def _serialize_results(self, results):
        serializer = getUtility(IBrainSerializer)
        return json.dumps(serializer.serialize_brains(results))

    def _count_unbatched_length(self, query):
        query = deepcopy(query)

        for key in ('sort_on', 'sort_order', 'sort_limit', 'batching_start'):
            if key in query:
                del query[key]

        catalog = getToolByName(self.context, 'portal_catalog')

        # ftw.solr may destroy our query
        # https://github.com/4teamwork/ftw.solr/issues/41
        query = deepcopy(query)
        brains = catalog(**query)
        return len(brains)



from ftw.bridge.client.interfaces import IWatcherPortletRegistry
from ftw.bridge.client.interfaces import PORTAL_URL_PLACEHOLDER
from ftw.bridge.client.testing import EXAMPLE_CONTENT_LAYER
from ftw.bridge.client.utils import json
from plone import api
from plone.uuid.interfaces import IUUID
from unittest2 import TestCase
from zope.component import queryMultiAdapter


DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S'


class TestWatcherFeedView(TestCase):

    layer = EXAMPLE_CONTENT_LAYER

    def test_component_is_registered(self):
        portal = self.layer['portal']
        request = self.layer['request']

        self.assertNotEquals(
            queryMultiAdapter((portal, request), name='watcher-feed'),
            None)

    def test_feed(self):
        portal = self.layer['portal']
        folder = self.layer['folder']
        page = folder.get('page')
        uid = IUUID(folder)

        request = self.layer['request']
        request.form['uid'] = uid

        view = queryMultiAdapter((portal, request), name='watcher-feed')
        result = view()
        self.assertEqual(type(result), str)

        data = json.loads(result)
        self.assertEqual(data, dict(
                title=u'Feed folder',
                details_url=u'%sfeed-folder/@@watcher-recently-modified' % (
                    PORTAL_URL_PLACEHOLDER),
                items=[
                    dict(title=u'The page with uml\xe4uts',
                         url=u'%sfeed-folder/page' % PORTAL_URL_PLACEHOLDER,
                         modified=page.modified().strftime(DATETIME_FORMAT),
                         portal_type=u'Document',
                         cssclass=u''),

                    dict(title=u'Feed folder',
                         url=u'%sfeed-folder' % PORTAL_URL_PLACEHOLDER,
                         modified=folder.modified().strftime(
                            DATETIME_FORMAT),
                         portal_type=u'Folder',
                         cssclass=u''),
                    ]))

    def test_ignore_types_in_feed(self):
        portal = self.layer['portal']
        folder = self.layer['folder']
        uid = IUUID(folder)

        request = self.layer['request']
        request.form['uid'] = uid

        view = queryMultiAdapter((portal, request), name='watcher-feed')

        data = json.loads(view())

        self.assertEqual(2, len(data.get('items')))

        api.portal.set_registry_record('types_to_ignore',
                                       value=[u'Document'],
                                       interface=IWatcherPortletRegistry)

        data = json.loads(view())

        self.assertEqual(1, len(data.get('items')))
        self.assertEqual('Folder', data.get('items')[0].get('portal_type'))



from ftw.bridge.client.interfaces import IBrainRepresentation
from ftw.bridge.client.interfaces import IBridgeRequest
from ftw.bridge.client.testing import EXAMPLE_CONTENT_LAYER
from ftw.builder import Builder
from ftw.builder import create
from plone.app.testing import TEST_USER_ID
from plone.app.testing import setRoles
from unittest2 import TestCase
from zope.component import getMultiAdapter
from zope.component import getUtility
import json


class TestBridgeSearchCatalogView(TestCase):

    layer = EXAMPLE_CONTENT_LAYER

    def setUp(self):
        self.portal = self.layer['portal']
        self.folder = self.layer['folder']
        self.request = self.layer['request']

    def test_component_registered(self):
        view = self.portal.restrictedTraverse('@@bridge-search-catalog')
        self.assertNotEqual(view, None)

    def test_view_returns_json(self):
        self.request.form['query'] = json.dumps({
                'path': '/'.join(self.folder.getPhysicalPath())})

        self.request.form['limit'] = 50

        view = getMultiAdapter((self.portal, self.request),
                               name='bridge-search-catalog')

        raw_data = view()
        data = json.loads(raw_data)

        self.assertEqual(len(data), 2)

        folder_data, page_data = data

        self.assertEqual(folder_data['_url'], u'***portal_url***feed-folder')
        self.assertEqual(folder_data['Title'], u'Feed folder')
        self.assertEqual(folder_data['id'], u'feed-folder')

        self.assertEqual(page_data['_url'],
                         u'***portal_url***feed-folder/page')
        self.assertEqual(page_data['Title'], u'The page with uml\xe4uts')
        self.assertEqual(page_data['id'], u'page')

    def test_count_unbatched_length(self):
        query = {'path': '/'.join(self.folder.getPhysicalPath()),
                 'sort_on': 'effective',
                 'sort_order': 'reverse',
                 'sort_limit': 1}

        view = getMultiAdapter((self.portal, self.request),
                               name='bridge-search-catalog')

        self.assertEqual(view._count_unbatched_length(query), 2)

    def test_response_contains_unbatched_length(self):
        self.request.form['query'] = json.dumps({
                'path': '/'.join(self.folder.getPhysicalPath()),
                'sort_on': 'effective',
                'sort_order': 'reverse',
                'sort_limit': 1})
        self.request.form['limit'] = 1

        view = getMultiAdapter((self.portal, self.request),
                               name='bridge-search-catalog')

        raw_data = view()
        data = json.loads(raw_data)
        self.assertEqual(len(data), 1)
        self.assertEqual(
            self.request.RESPONSE.getHeader('X-total_results_length'),
            '2')


class TestCatalogRequest(TestCase):

    layer = EXAMPLE_CONTENT_LAYER

    def setUp(self):
        self.folder = self.layer['folder']

    def test_full_stack_catalog_query_by_traversing(self):
        query = {
            'path': '/'.join(self.folder.getPhysicalPath())}

        utility = getUtility(IBridgeRequest)
        results = utility.search_catalog('current-client', query)

        self.assertEqual(len(results), 2)
        folder, page = results

        self.assertTrue(IBrainRepresentation.providedBy(folder))
        self.assertTrue(IBrainRepresentation.providedBy(page))

        self.assertEqual(folder.Title, 'Feed folder')
        self.assertEqual(page.Title, 'The page with uml\xc3\xa4uts')

        self.assertEqual(folder.getURL(), 'http://nohost/plone/feed-folder')
        self.assertEqual(page.getURL(),
                         'http://nohost/plone/feed-folder/page')

        self.assertEqual(folder.portal_type, 'Folder')
        self.assertEqual(page.portal_type, 'Document')

    def test_batched_results_with_respons_headers(self):
        query = {
            'path': '/'.join(self.folder.getPhysicalPath()),
            'sort_on': 'sortable_title',
            'sort_limit': 1}

        utility = getUtility(IBridgeRequest)
        results = utility.search_catalog('current-client', query)

        self.assertEqual(len(results), 1)
        folder = results[0]

        self.assertTrue(IBrainRepresentation.providedBy(folder))
        self.assertEqual(folder.Title, 'Feed folder')

    def test_reverse_sorting(self):
        setRoles(self.layer['portal'], TEST_USER_ID, ['Manager'])
        foo = create(Builder('folder').titled('Foo'))
        create(Builder('folder').titled('Bar').within(foo))

        query = {'path': '/'.join(foo.getPhysicalPath()),
                 'sort_on': 'sortable_title',
                 'sort_order': 'reverse'}
        utility = getUtility(IBridgeRequest)
        results = utility.search_catalog('current-client', query)
        titles = map(lambda brain: brain.Title, results)

        self.assertEquals(['Foo', 'Bar'], titles)



from Products.CMFCore.utils import getToolByName
from Products.Five import BrowserView
from Products.statusmessages.interfaces import IStatusMessage
from ftw.bridge.client import _
from ftw.bridge.client.exceptions import MaintenanceError
from ftw.bridge.client.interfaces import IBridgeRequest
from ftw.bridge.client.interfaces import MAINTENANCE_ERROR_MESSAGE
from ftw.bridge.client.interfaces import PORTAL_URL_PLACEHOLDER
from zope.component import getMultiAdapter
from zope.component import getUtility


class AddFavorite(BrowserView):
    """This view adds a Link object to the "Favorites" folder located in
    the home folder of the current member. The Link has `title` and is
    pointing to the `url` passed in as request data.
    The view is meant to be called on the site hosting the dashboard and
    can be called by the bridge proxy.
    """

    def __call__(self):
        url = self.request.form['url']
        title = self.request.form['title']

        folder = self.get_favorites_folder()
        id_ = self.generate_favorite_id(folder)
        folder.invokeFactory('Link', id=id_, title=title, remoteUrl=url)

        return 'OK'

    def get_favorites_folder(self):
        membership = getToolByName(self.context, 'portal_membership')
        folder = membership.getHomeFolder()
        if 'Favorites' not in folder.objectIds():
            folder.invokeFactory('Folder', 'Favorites', title='Favorites')
        return folder.get('Favorites')

    def generate_favorite_id(self, favorites_folder, base='favorite'):
        existing_ids = favorites_folder.objectIds()

        idx = 1
        while True:
            id_ = '%s-%d' % (base, idx)
            if id_ not in existing_ids:
                return id_
            idx += 1


class RemoteAddFavoriteAction(BrowserView):
    """This view adds a favorite (Link object) on the remote site called
    "dashboard" over the bridge proxy. The link object points to the current
    context.
    """

    def __call__(self):
        self._create_favorite(title=self.context.Title(),
                              url=self._get_url())

        referer = self.request.environ.get('HTTP_REFERER')
        if referer:
            self.request.RESPONSE.redirect(referer)
        else:
            state = getMultiAdapter((self.context, self.request),
                                    name='plone_context_state')
            self.request.RESPONSE.redirect(state.view_url())

    def _create_favorite(self, title, url):
        data = {'title': title,
                'url': url}

        requester = getUtility(IBridgeRequest)
        try:
            response = requester('dashboard', '@@add-favorite',
                                 data=data, silent=True)

        except MaintenanceError:
            IStatusMessage(self.request).addStatusMessage(
                MAINTENANCE_ERROR_MESSAGE, type='error')
            return False

        if response is None or response.code != 200:
            IStatusMessage(self.request).addStatusMessage(
                _(u'favorite_creation_failed',
                  default=u'The favorite could not be created.'),
                type='error')
            return False

        else:
            IStatusMessage(self.request).addStatusMessage(
                _(u'info_favorite_created',
                  default=u'${title} was added to your favorites.',
                  mapping={'title': title.decode('utf-8')}),
                type='info')
            return True

    def _get_url(self):
        state = getMultiAdapter((self.context, self.request),
                                name='plone_context_state')

        portal_url = getToolByName(self.context, 'portal_url')() + '/'
        relative_path = state.view_url()[len(portal_url):]
        return ''.join((PORTAL_URL_PLACEHOLDER, relative_path))



from ftw.bridge.client.tabbed import interfaces
from ftw.bridge.client.tabbed.components import BridgeCatalogTableSource
from ftw.bridge.client.testing import BRIDGE_CONFIG_LAYER
from ftw.bridge.client.tests.base import RequestAwareTestCase
from ftw.table.interfaces import ITableSource
from zope.component import queryMultiAdapter
import json


class TestTableSource(RequestAwareTestCase):

    layer = BRIDGE_CONFIG_LAYER

    def test_component_registered(self):
        config = self.providing_stub(
            interfaces.IBridgeCatalogTableSourceConfig)
        request = self.create_dummy()

        self.replay()

        component = queryMultiAdapter((config, request), ITableSource)
        self.assertNotEqual(component, None)
        self.assertEqual(type(component), BridgeCatalogTableSource)

    def test_implements_interface(self):
        self.replay()
        self.assertTrue(ITableSource.implementedBy(BridgeCatalogTableSource))

    def test_search_results_requests_bridge(self):
        config = self.providing_stub(
            interfaces.IBridgeCatalogTableSourceConfig)
        self.expect(config.bridge_remote_client_id).result('other-client')
        self.expect(config.bridge_remove_path).result(False)
        self.expect(config.batching_pagesize).result(50)
        self.expect(config.pagesize).result(50)
        self.expect(config.batching_current_page).result(1)
        request = self.create_dummy(
            RESPONSE=self.create_dummy(
                headers={}))

        response_data = [
            {'Title': 'Foo',
             '_url': 'https://target-client/foo',
             'portal_type': 'Folder'},
            {'Title': 'Bar',
             '_url': 'https://target-client/bar',
             'portal_type': 'Folder'}]

        query = {'portal_type': ['Folder']}

        response = self._create_response(raw=json.dumps(response_data))
        url = 'http://bridge/proxy/other-client/@@bridge-search-catalog'

        self._expect_request(
            url=url,
            data={'query': json.dumps({'portal_type': ['Folder'],
                                       'batching_start': 0}),
                  'limit': 50}).result(response)

        self.replay()
        component = queryMultiAdapter((config, request), ITableSource)
        results = component.search_results(query)
        self.assertEqual(len(results), 2)

    def test_search_results_fails_when_wrong_configured(self):
        config = self.providing_stub(
            interfaces.IBridgeCatalogTableSourceConfig)
        self.expect(config.bridge_remote_client_id).result(None)
        request = self.create_dummy()

        self.replay()
        component = queryMultiAdapter((config, request), ITableSource)
        with self.assertRaises(ValueError) as cm:
            component.search_results({})
        self.assertIn(
            'defines no bridge_remote_client_id',
            str(cm.exception))

    def test_path_removed(self):
        config = self.providing_stub(
            interfaces.IBridgeCatalogTableSourceConfig)
        self.expect(config.bridge_remote_client_id).result('other-client')
        self.expect(config.bridge_remove_path).result(True)
        self.expect(config.batching_pagesize).result(50)
        self.expect(config.pagesize).result(50)
        self.expect(config.batching_current_page).result(1)
        request = self.create_dummy()

        query = {'portal_type': ['Folder'],
                 'path': '/Plone'}

        response = self._create_response(raw='[]')
        url = 'http://bridge/proxy/other-client/@@bridge-search-catalog'

        self._expect_request(
            url=url,
            data={'query': json.dumps({'portal_type': ['Folder'],
                                       'batching_start': 0}),
                  'limit': 50}).result(response)

        self.replay()
        component = queryMultiAdapter((config, request), ITableSource)
        component.search_results(query)



from ftw.bridge.client.testing import EXAMPLE_CONTENT_LAYER
from unittest2 import TestCase
import os.path


class TestRecentlyModifiedView(TestCase):

    layer = EXAMPLE_CONTENT_LAYER

    def setUp(self):
        self.folder = self.layer['portal'].get('feed-folder')
        self.layer['request'].ACTUAL_URL = os.path.join(
            self.folder.absolute_url(),
            '@@watcher-recently-modified')

    def test_batched_results(self):
        view = self.folder.restrictedTraverse('@@watcher-recently-modified')
        view.update()
        result = view.batched_results

        self.assertNotEqual(result, None)
        self.assertEqual(result.__class__.__name__, 'Batch')
        self.assertEqual(len(result), 2)

        # do not expect a order since the folder and the page are
        # created at the same time - it may switch in some cases.
        brain_ids = [brain.id for brain in result]
        self.assertEqual(set(brain_ids), set(['feed-folder', 'page']))

    def test_template_renders(self):
        view = self.folder.restrictedTraverse('@@watcher-recently-modified')
        html = view()

        self.assertIn(u'The page with uml\xe4uts', html)
        self.assertIn(u'Feed folder', html)



from ftw.bridge.client.interfaces import PORTAL_URL_PLACEHOLDER
from Products.CMFCore.utils import getToolByName
from zope.component.hooks import getSite

try:
    import json
except ImportError:
    import simplejson as json


def get_object_url(obj):
    """Returns the url to this object using, replacing the site url with a
    placeholder replaced with the public URL by the bridge.
    """
    portal_url = getToolByName(obj, 'portal_url')() + '/'
    return obj.absolute_url().replace(portal_url, PORTAL_URL_PLACEHOLDER)


def get_brain_url(brain):
    """Returns the brain to this object using, replacing the site url with a
    placeholder replaced with the public URL by the bridge.
    """
    portal_url = getToolByName(getSite(), 'portal_url')() + '/'
    return brain.getURL().replace(portal_url, PORTAL_URL_PLACEHOLDER)


def to_utf8_recursively(data):
    if isinstance(data, unicode):
        return data.encode('utf-8')
    elif isinstance(data, list):
        return map(to_utf8_recursively, data)
    elif isinstance(data, dict):
        return dict(map(lambda item: map(to_utf8_recursively, item),
                        data.items()))
    else:
        return data



from Products.PluggableAuthService.interfaces import plugins
from copy import copy
from ftw.bridge.client.interfaces import IBridgeConfig
from ftw.bridge.client.plugin import BridgePlugin
from ftw.bridge.client.testing import ZCML_LAYER
from ftw.testing import MockTestCase
from mocker import ARGS, MATCH
from zope.browser.interfaces import IAdding
from zope.component import queryMultiAdapter
from zope.interface.verify import verifyClass


class TestAddBridgePlugin(MockTestCase):

    layer = ZCML_LAYER

    def test_component_is_registered(self):
        adding = self.providing_stub(IAdding)
        request = self.stub_request()

        self.replay()

        addview = queryMultiAdapter((adding, request),
                                    name='add-bridge-plugin')
        self.assertNotEqual(addview, None)

    def test_addview_adds_plugin(self):
        context = self.mocker.mock()
        adding = self.providing_stub(IAdding)
        request = self.stub_request(stub_response=False)

        self.expect(request.form).result({
                'form.button.Add': 'Add',
                'id': 'the-bridge',
                'title': 'The Bridge'})

        self.expect(adding.context).result(context)
        context['the-bridge'] = MATCH(
            lambda obj: isinstance(obj, BridgePlugin))

        self.expect(adding.absolute_url()).result('http://nohost/acl_users')
        self.expect(request.response.redirect(
                'http://nohost/acl_users/manage_workspace?' + \
                    'manage_tabs_message=Plugin+added.'))

        self.replay()

        addview = queryMultiAdapter((adding, request),
                                    name='add-bridge-plugin')
        addview()


class TestPasPlugin(MockTestCase):

    layer = ZCML_LAYER

    def setUp(self):
        MockTestCase.setUp(self)
        self.config = self.stub_interface(IBridgeConfig)
        self.expect(self.config.get_bridge_ips()).result([
                '127.0.0.1', '127.0.0.2'])
        self.mock_utility(self.config, IBridgeConfig)

        self.valid_credentials = {'id': 'john.doe',
                                  'login': 'john.doe',
                                  'ip': '127.0.0.1',
                                  'extractor': 'bridge'}

    def test_implements_interfaces(self):
        self.replay()
        self.assertTrue(plugins.IExtractionPlugin.implementedBy(BridgePlugin))
        verifyClass(plugins.IExtractionPlugin, BridgePlugin)

        self.assertTrue(plugins.IAuthenticationPlugin.implementedBy(
                BridgePlugin))
        verifyClass(plugins.IAuthenticationPlugin, BridgePlugin)

    def test_extracts_credentials(self):
        request = self.stub_request()
        headers = {'X-BRIDGE-AC': 'john.doe',
                   'X-BRIDGE-ORIGIN': 'client-one'}
        self.expect(request.get_header(ARGS)).call(
            lambda key, default: headers.get(key, default))
        self.expect(request.environ).result({
                'REMOTE_ADDR': '127.0.0.1'})

        self.replay()

        plugin = BridgePlugin('bridge')
        self.assertEqual(
            plugin.extractCredentials(request),
            {'id': 'john.doe',
             'login': 'john.doe',
             'ip': '127.0.0.1'})

    def test_extract_credentials_requires_origin(self):
        request = self.stub_request()
        headers = {'X-BRIDGE-AC': 'john.doe',
                   'X-BRIDGE-ORIGIN': None}
        self.expect(request.get_header(ARGS)).call(
            lambda key, default: headers.get(key, default))
        self.expect(request.environ).result({})

        self.replay()

        plugin = BridgePlugin('bridge')
        self.assertEqual(
            plugin.extractCredentials(request),
            {})

    def test_extract_credentials_requires_principal(self):
        request = self.stub_request()
        headers = {'X-BRIDGE-AC': None,
                   'X-BRIDGE-ORIGIN': 'client-one'}
        self.expect(request.get_header(ARGS)).call(
            lambda key, default: headers.get(key, default))
        self.expect(request.environ).result({})

        self.replay()

        plugin = BridgePlugin('bridge')
        self.assertEqual(
            plugin.extractCredentials(request),
            {})

    def test_authenticate_credentials(self):
        self.replay()
        plugin = BridgePlugin('bridge')
        plugin.REQUEST = self.create_dummy()

        self.assertEqual(
            plugin.authenticateCredentials(self.valid_credentials),
            ('john.doe', 'john.doe'))

    def test_authenticate_credentials_validates_extractor(self):
        self.replay()
        plugin = BridgePlugin('bridge')
        plugin.REQUEST = self.create_dummy()

        creds = copy(self.valid_credentials)
        creds['extractor'] = 'wrong-extractor'

        self.assertEqual(
            plugin.authenticateCredentials(creds),
            None)

    def test_authenticate_credentials_from_bad_ip_fails(self):
        self.replay()
        plugin = BridgePlugin('bridge')
        plugin.REQUEST = self.create_dummy()

        creds = copy(self.valid_credentials)
        creds['ip'] = '192.168.1.2'

        self.assertEqual(
            plugin.authenticateCredentials(creds),
            None)

    def test_authenticate_credentials_from_right_ip_succeeds(self):
        self.replay()
        plugin = BridgePlugin('bridge')
        plugin.REQUEST = self.create_dummy()

        creds = copy(self.valid_credentials)
        creds['ip'] = '127.0.0.2'

        self.assertEqual(
            plugin.authenticateCredentials(creds),
            ('john.doe', 'john.doe'))

    def test_get_request_ip(self):
        self.replay()
        plugin = BridgePlugin('bridge')

        request = self.create_dummy(environ={})
        self.assertEqual(plugin._get_request_ip(request), None)

        request = self.create_dummy(environ={'REMOTE_ADDR': '127.0.0.1'})
        self.assertEqual(plugin._get_request_ip(request), '127.0.0.1')

        request = self.create_dummy(environ={
                'REMOTE_ADDR': '127.0.0.1',
                'HTTP_X_FORWARDED_FOR': '192.168.1.1'})
        self.assertEqual(plugin._get_request_ip(request), '127.0.0.1')

        request = self.create_dummy(environ={
                'REMOTE_ADDR': '127.0.0.1',
                'HTTP_X_FORWARDED_FOR': ' 192.168.1.1 , 10.10.10.10'})
        self.assertEqual(plugin._get_request_ip(request), '127.0.0.1')



from ftw.upgrade import UpgradeStep


class AddRegistryEntryToFilterWatcherFeed(UpgradeStep):
    """Add registry entry to filter watcher feed.
    """

    def __call__(self):
        self.install_upgrade_profile()



from setuptools import setup, find_packages
import os

version = '1.0.13'

tests_require = [
    'ftw.builder',
    'ftw.testing',
    'mocker',
    'plone.app.testing',
    'plone.mocktestcase',
    'plone.testing',
    'plone.uuid',
    'transaction',
    'unittest2',
    'z3c.autoinclude',  # TODO: eliminate autoinclude in tests
    'zope.browser',
    ]

extras_require = {
    'tests': tests_require,
    'tabbedview': [
        'ftw.tabbedview>=3.2.3',
        'ftw.table']}

tests_require += extras_require['tabbedview']


setup(name='ftw.bridge.client',
      version=version,
      description='Adds ftw.bridge support to plone.',

      long_description=open('README.rst').read() + '\n' + \
          open(os.path.join('docs', 'HISTORY.txt')).read(),

      # Get more strings from
      # http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
        'Framework :: Plone',
        'Framework :: Plone :: 4.1',
        'Framework :: Plone :: 4.2',
        'Framework :: Plone :: 4.3',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python',
        ],

      keywords='ftw bridge proxy client',
      author='4teamwork AG',
      author_email='mailto:info@4teamwork.ch',
      url='https://github.com/4teamwork/ftw.bridge.client',
      license='GPL2',

      packages=find_packages(),
      namespace_packages=['ftw', 'ftw.bridge'],
      include_package_data=True,
      zip_safe=False,

      install_requires=[
          'AccessControl',
          'Acquisition',
          'Missing',
          'Plone',
          'Products.CMFCore',
          'Products.CMFPlone',
          'Products.GenericSetup',
          'Products.PluggableAuthService',
          'Products.statusmessages',
          'Zope2',
          'ftw.profilehook',
          'ftw.upgrade',
          'plone.api>=1.4.11',
          'plone.app.portlets',
          'plone.app.registry>=1.2a1',
          'plone.memoize',
          'plone.portlets',
          'setuptools',
          'zope.app.component',
          'zope.component',
          'zope.formlib',
          'zope.i18nmessageid',
          'zope.interface',
          'zope.publisher',
          'zope.schema',
      ],

      tests_require=tests_require,
      extras_require=extras_require,

      entry_points='''
      [z3c.autoinclude.plugin]
      target = plone
      ''',
      )



from ftw.testing.genericsetup import apply_generic_setup_layer
from ftw.testing.genericsetup import GenericSetupUninstallMixin
from unittest2 import TestCase


@apply_generic_setup_layer
class TestGenericSetupUninstall(TestCase, GenericSetupUninstallMixin):
    package = 'ftw.bridge.client'



# pylint: disable=E0211, E0213
# E0211: Method has no argument
# E0213: Method should have "self" as first argument

from ftw.bridge.client import _
from zope import schema
from zope.interface import Interface


PORTAL_URL_PLACEHOLDER = '***portal_url***'


MAINTENANCE_ERROR_MESSAGE = _(
    u'error_msg_maintenance',
    default=u'The target service is currently in maintenace. Try again later.')


class IBridgeConfig(Interface):
    """Utility providing bridge configuration infos. Configurable using
    environment variables:

    ``bridge_url`` -- Full url to the bridge /proxy view.
    ``bridge_ips`` -- A comma seperated string of allowed bridge IPs.
    ``bridge_client_id`` -- The id (string) of this client.
    """

    def get_url():
        """Returns the URL to the proxy view of the bridge, ending with a
        slash (/).
        """

    def get_bridge_ips():
        """Returns a list of allowed IP addresses.
        """

    def get_client_id():
        """Returns the client id of this client.
        """


class IBridgeRequest(Interface):
    """A utility for sending requests to the bridge.
    """

    def __call__(target, path, headers=None, data=None,
                 silent=False):
        """Make a request to the client ``target`` client through the
        configured bridge.

        Arguments:
        ``target`` -- Target client id as configured on the bridge.
        ``path`` -- Path and view to request, relative to the site root on
        the remote client.
        ``headers`` -- Dict of additional request headers (optional).
        ``data`` -- Dict of data to pass (optional).
        ``silent`` -- Return None on connection error instead of raising
        an exception. The original exception will be added to the error_log.
        """

    def get_json(*args, **kwargs):
        """Makes a request to a view returning json and converts the
        json to python.

        The arguments are the same as for the __call__ method.
        """

    def search_catalog(target, query, limit=50):
        """Search the remote catalog and return a list of
        ``IBrainRepresentation`` objects.

        Arguments:
        ``target`` -- Target client id as configured on the bridge.
        ``query`` -- A jsonizable dict with the catalog query.
        ``limit`` -- Limits the remote result to this number of brains for
        avoiding huge responses. If necessary this can be changed. Defaults
        to 50.
        """


class IBrainSerializer(Interface):
    """A tool for serializing and deserializing a set of brains, so that
    it can be transported to another client.
    The deserialized brains are represented with ``IBrainRepresentation``
    objects providing the original metadata. Those representation objects
    cannot retrieve the object (since the object is on another client).
    """

    def serialize_brains(results):
        """Serialize a list of brains into a jsonizable list.
        """

    def deserialize_brains(data):
        """Deserialize the previously serialized data to
        ``IBrainRepresentation`` objects.
        """


class IBrainRepresentation(Interface):
    """Represents a catalog brain which was transported over the bridge.
    It contains all known remote metadata and some important methods.
    """

    def __init__(data):
        """Creates a brain representation instance by passing in a dict
        created previously be the ``brain_to_data`` method.
        """

    def getURL():
        """Returns the full public url to the object on the remote client.
        """


class IWatcherPortletRegistry(Interface):
    """Describe some properties to configure the watcher portlet.
    """

    types_to_ignore = schema.List(
        title=_(u"Types to ignore"),
        description=_(
            u"Defined portaltypes will be ignorded by the watcher portlet"),
        required=False,
        value_type=schema.TextLine(title=u"Portal types"),
        default=[]
    )



from Products.CMFCore.utils import getToolByName
from ftw.bridge.client.browser import watcher
from ftw.bridge.client.exceptions import MaintenanceError
from ftw.bridge.client.interfaces import PORTAL_URL_PLACEHOLDER
from ftw.bridge.client.portlets.watcher import IWatcherPortlet
from ftw.bridge.client.testing import EXAMPLE_CONTENT_LAYER
from ftw.bridge.client.tests.base import RequestAwareTestCase
from ftw.bridge.client.utils import json
from plone.portlets.constants import USER_CATEGORY
from plone.portlets.interfaces import IPortletManager
from plone.portlets.utils import hashPortletInfo
from plone.uuid.interfaces import IUUID
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.component import queryMultiAdapter


DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S'


class TestAjaxLoadPortletDataView(RequestAwareTestCase):

    layer = EXAMPLE_CONTENT_LAYER

    def _get_portlet_and_hash(self):
        context = self.layer['portal']
        column_manager = getUtility(IPortletManager, name='plone.dashboard1')
        membership_tool = getToolByName(context, 'portal_membership')
        userid = membership_tool.getAuthenticatedMember().getId()
        user_category = column_manager.get(USER_CATEGORY, None)
        assert user_category
        column = user_category.get(userid, None)
        assert column

        for portlet in column.values():
            if IWatcherPortlet.providedBy(portlet):
                info = {'manager': 'plone.dashboard1',
                        'category': USER_CATEGORY,
                        'key': userid,
                        'name': portlet.getId()}
                return portlet, hashPortletInfo(info)

    def setUp(self):
        RequestAwareTestCase.setUp(self)

        portal = self.layer['portal']
        request = self.layer['request']

        request.environ['HTTP_X_BRIDGE_ORIGIN'] = 'client-one'
        request.form['path'] = '@@watcher-feed?uid=567891234'

        self.assertEqual(watcher.AddWatcherPortlet(portal, request)(), 'OK')

    def test_view(self):
        portal = self.layer['portal']
        request = self.layer['request']
        folder = self.layer['folder']
        page = folder.get('page')

        request.form['uid'] = IUUID(folder)
        feed_view = getMultiAdapter((portal, request), name='watcher-feed')
        feed_data = feed_view()

        portlet, portlet_hash = self._get_portlet_and_hash()
        url = 'http://bridge/proxy/%s/%s' % (portlet.client_id,
                                             portlet.path)

        self._expect_request(url=url).result(self._create_response(
                status_code=200, raw=feed_data))

        self.replay()
        request.form['hash'] = portlet_hash

        view = queryMultiAdapter((portal, request), name='watcher-load-data')
        self.assertNotEqual(view, None)

        dt_format = '%b %d, %Y'

        expected_feed_data = dict(
            title=u'Feed folder',
            details_url=u'%sfeed-folder/@@watcher-recently-modified' % (
                PORTAL_URL_PLACEHOLDER),
            items=[
                dict(title=u'The page with uml\xe4uts',
                     url=u'%sfeed-folder/page' % PORTAL_URL_PLACEHOLDER,
                     modified=page.modified().strftime(dt_format),
                     portal_type=u'Document',
                     cssclass=u''),

                dict(title=u'Feed folder',
                     url=u'%sfeed-folder' % PORTAL_URL_PLACEHOLDER,
                     modified=folder.modified().strftime(dt_format),
                     portal_type=u'Folder',
                     cssclass=u''),
                ])

        self.assertEqual(json.loads(view()), expected_feed_data)

    def test_maintenance_error(self):
        portal = self.layer['portal']
        request = self.layer['request']
        folder = self.layer['folder']

        request.form['uid'] = IUUID(folder)

        _portlet, portlet_hash = self._get_portlet_and_hash()
        self._expect_request().throw(MaintenanceError())

        self.replay()
        request.form['hash'] = portlet_hash

        view = queryMultiAdapter((portal, request), name='watcher-load-data')
        self.assertNotEqual(view, None)

        self.assertEqual(view(), '"MAINTENANCE"')

    def test_view_does_not_fail_when_no_modified_date_is_passed(self):
        portal = self.layer['portal']
        request = self.layer['request']
        folder = self.layer['folder']

        request.form['uid'] = IUUID(folder)
        feed_view = getMultiAdapter((portal, request), name='watcher-feed')
        feed_data = feed_view()

        # modify feed data
        data = json.loads(feed_data)
        for item in data['items']:
            item['modified'] = ''
        data = json.dumps(data)

        _portlet, portlet_hash = self._get_portlet_and_hash()
        self._expect_request().result(self._create_response(
                status_code=200, raw=data))

        self.replay()
        request.form['hash'] = portlet_hash

        view = queryMultiAdapter((portal, request), name='watcher-load-data')
        self.assertNotEqual(view, None)
        self.assertEqual(type(json.loads(view())), dict)
